# To be added to `lambeq/ansatz/circuit.py`


def _sim_ansatz_factory(sim_circuit_n=14):
    """Generate a SimAnsatz class for the specified Sim ansatz.

    Parameters
    ----------
    sim_circuit_n : int
        Circuit number, according to arXiv:1905.10876

    """

    # Map from Sim circuit number to a function specifying
    # number of params, given the number of qubits
    param_count_fns = {14: lambda n_qbs: 4 * n_qbs,
                       15: lambda n_qbs: 2 * n_qbs}

    sim_discopy_classes = {14: Sim14, 15: Sim15}

    if sim_circuit_n not in sim_discopy_classes:
        raise ValueError(f'Invalid Sim circuit name: {sim_circuit_n}')

    class SimAnsatz(CircuitAnsatz):
        """ Sim circuit ansatz (arXiv:1905.10876)

        Multiple Sim ansatze can be created, using the available
         DisCoPy classes specified in `sim_discopy_classes`.

        """

        def __init__(self,
                     ob_map: Mapping[Ty, int],
                     n_layers: int,
                     n_single_qubit_params: int = 3,
                     discard: bool = False,
                     special_cases: Optional[Callable[[_ArMapT],
                                                      _ArMapT]] = None):
            """Instantiate Sim circuit ansatz (arXiv:1905.10876).

            Parameters
            ----------
            ob_map : dict
                A mapping from :py:class:`discopy.rigid.Ty` to the number of
                qubits it uses in a circuit.
            n_layers : int
                The number of IQP layers used by the ansatz.
            n_single_qubit_params : int, default: 3
                The number of single qubit rotations used by the ansatz.
            discard : bool, default: False
                Discard open wires instead of post-selecting.
            special_cases : callable, optional
                A function that transforms an arrow map into one specifying
                special cases that should not be converted by the Ansatz
                class.

            """

            super().__init__(ob_map=ob_map, n_layers=n_layers,
                             n_single_qubit_params=n_single_qubit_params)

            if special_cases is None:
                special_cases = self._special_cases

            self.n_layers = n_layers
            self.n_single_qubit_params = n_single_qubit_params
            self.discard = discard
            self.functor = Functor(ob=self.ob_map,
                                   ar=special_cases(self._ar))

        def _ar(self, box: Box) -> Circuit:
            label = self._summarise_box(box)
            dom, cod = self._ob(box.dom), self._ob(box.cod)

            n_qubits = max(dom, cod)
            n_layers = self.n_layers
            n_1qubit_params = self.n_single_qubit_params

            if n_qubits == 0:
                circuit = Id()
            elif n_qubits == 1:
                syms = [Symbol(f'{label}_{i}') for i in range(n_1qubit_params)]
                rots = [Rx, Rz]
                circuit = Id(qubit)
                for i, sym in enumerate(syms):
                    circuit >>= rots[i % 2](sym)
            else:
                n_params = n_layers * param_count_fns[sim_circuit_n](n_qubits)
                syms = [Symbol(f'{label}_{i}') for i in range(n_params)]
                params = np.array(syms).reshape(
                        (n_layers, param_count_fns[sim_circuit_n](n_qubits)))

                circuit = sim_discopy_classes[sim_circuit_n](n_qubits, params)

            if cod > dom:
                circuit <<= Id(dom) @ Ket(*[0]*(cod - dom))
            elif self.discard:
                circuit >>= Id(cod) @ Discard(dom - cod)
            else:
                circuit >>= Id(cod) @ Bra(*[0]*(dom - cod))
            return circuit

    return SimAnsatz


Sim14Ansatz = _sim_ansatz_factory(14)
Sim15Ansatz = _sim_ansatz_factory(15)
